import streamlit as st
from tw_experimentation.streamlit.streamlit_utils import (
    _coming_from_other_page,
    swap_evaluation_change_alpha,
    initalise_session_states,
    swap_checkbox_state,
    frequentist_results,
    cuped_results,
    segmentation_analysis,
)

import numpy as np
from tw_experimentation.utils import highlight


def page_4_experiment_evaluation():
    # results

    st.session_state.update(st.session_state)

    CURRENT_PAGE = "Experiment Evaluation"
    initalise_session_states()

    st.header("Experiment evaluation")

    if st.session_state["is_defined_data_model"] and st.session_state["is_experiment"]:
        if st.session_state["output_loaded"] is not None:
            emd = st.session_state["output_loaded"]["emd"]
        else:
            emd = st.session_state.ed.experiment_meta_data()

        frequentist_output = frequentist_results(
            st.session_state.ed, _output_loaded=st.session_state["output_loaded"]
        )

        st.markdown(
            """
            - In this section you can test hypotheses about your groups
            - And check if you have a statistically significant result by getting the p-value
            """
        )

        st.session_state["evaluation_change_alpha"] = st.checkbox(
            "Change significance level (defaults to .05)",
            value=st.session_state["evaluation_change_alpha"],
            on_change=swap_evaluation_change_alpha,
        )

        if st.session_state["evaluation_change_alpha"]:
            st.session_state["evaluation_alpha"] = st.slider(
                r"significance level $ \alpha $ (Type-I error)",
                min_value=0.001,
                max_value=100.0,
                format="%.2f%%",
                # key="evaluation_alpha",
                value=(
                    st.session_state["evaluation_alpha"]
                    if _coming_from_other_page(
                        CURRENT_PAGE, st.session_state["last_page"]
                    )
                    else 5.0
                ),
                step=0.01,
            )
        else:
            st.session_state["evaluation_alpha"] = 5.0
        alpha = st.session_state.evaluation_alpha
        st.markdown(
            """
            In case you have multiple treatment variants, you will have the option to perform a
            multi-test correction on the p-values
            (current implementation: Bonferroni method. This is subject to change in the future).
            """
        )

        if emd.n_variants > 2:
            st.session_state["multi-test-correction"] = st.radio(
                "Multi-test correction:", ["Yes", "No"]
            )

        st.subheader("Frequentist Evaluation")
        multitest_correction = (
            "bonferroni"
            if emd.n_variants > 2 and st.session_state["multi-test-correction"] == "Yes"
            else None
        )

        st.dataframe(
            frequentist_output.compute_stats_per_target(
                type_i_error=alpha / 100, multitest_correction=multitest_correction
            )
            .get_results_table()
            .style.apply(highlight, axis=1)
            .set_table_attributes("style='display:inline'")
            .bar(subset=["Estimated_Effect_relative"], color="grey")
            .format(precision=3),
        )

        st.markdown(
            """
            #### How can I interpret the results?

            - We compare each treatment variant with the control group
            - The statistic of interest is the difference in means for every specified metric.
            - This yields estimated treatment effects and p-value for them
            - By convention, if p-value < 0.05 the result is regarded significant
            - We also provide confidence intervals for metrics
            """
        )
        st.divider()

        if (
            emd.pre_experiment_cols is not None
            and emd.pre_experiment_cols != []
            and emd.pre_experiment_cols != [None]
        ):  # TODO: could probably be done better
            st.session_state["evaluate_CUPED"] = st.toggle(
                "Power/Sensitivity improvement via CUPED",
                value=(
                    st.session_state["evaluate_CUPED"]
                    if "evaluate_CUPED" in st.session_state
                    else False
                ),
                on_change=swap_checkbox_state,
                args=("evaluate_CUPED",),
                help="""
            CUPED is a variance reduction method leveraging pre-experiment data in order to
            increase the sensitivity of an A/B test.
            The basic idea is to use pre-experiment data as a control variate in the test;  the pre-experiment data is used to
            transform the target variable so that its variability is lowered
            after which we apply the standard/vanilla T-test to the transformed target.
            """,
            )
            if st.session_state["evaluate_CUPED"]:
                st.markdown(
                    """
                    The columns specified as pre-experiment metrics in Data Loading must be independent of the variant assignment.
                    The greater the correlation between the chosen columns and the outcome, the greater the CUPED's variance reduction and hence power/sensitivity improvement.
                    **The simplest and best option is to use the pre-experiment outcome if such exists**.
                    For example, if the outcome metric is volume of transactions, then the pre-experiment
                    outcome could be the volume of transactions in the month prior to the experiment.
                    """
                )
                st.subheader("Test with power/sensitivity improvement")
                cuped_output = cuped_results(
                    st.session_state.ed,
                    _output_loaded=st.session_state["output_loaded"],
                )

                st.dataframe(
                    cuped_output.results_table(
                        has_correction=(
                            emd.n_variants > 2
                            and st.session_state["multi-test-correction"] == "Yes"
                        ),
                        alpha=alpha / 100,
                    )
                    .style.set_table_attributes(  # .apply(highlight, axis=1)\
                        "style='display:inline'"
                    )  # .bar(subset=["Estimated_Effect_relative"], color="grey")
                    .format(precision=3)
                )

                st.markdown(
                    """
                    #### How can I interpret the results?
                    - The above presents the result of the T-test after
                    the power/sensitivity method has been applied
                    - The conclusions which can be drawn are the same
                    as for the vanilla T-tests

                    #### What if the CUPED's effect significance is different from the Frequentist Evaluation's effect significance?
                    - Generally, we can still consider the effect to be significant if CUPED
                    is significant while Frequentist Evaluation is not.
                    - However, if it's the other way and CUPED's effect is not significant while Frequentist Evaluation's is,
                    we have a reason to believe that there is some issue with CUPED that requires further investigation.
                    """
                )
                st.divider()
        st.subheader("Segmentation with Wise Pizza")
        st.markdown(
            """
            - In this section you can find unusual segments in terms of the
            difference between the control and test groups
            - Please provide segments which you want to analyse, metric to
            analyse and number of observations
            """
        )

        st.markdown(
            """
            Find segments whose average is most different from the global one

            - `segments`: List of discrete dimensions to find slices
            - `target`: Metric to analyse
            - `treatment`: If you have different test groups, specify group here,
            for example treatment=1 means we compare with first treatment group
            - `min_segments`: Minimum number of segments to find
            - `max_depth`: Maximum number of dimension to constrain in
            segment definition


            *Warning*: The p-values are currently not corrected for
            multiple comparisons.
            However, Wise-Pizza identifies segments as interesting only
            if the treatment effect is sufficiently high
            compared to the segment sample size so this selection is a
            first approximation
            of avoiding p-value inflation in segmentation analysis.
            """
        )

        if (
            st.session_state["segments"] is None
            or len(st.session_state["segments"]) == 0
        ):
            st.write("No segments available for wise pizza segmentation analysis")
        else:
            segmentation_output = segmentation_analysis(
                st.session_state.ed, _output_loaded=st.session_state["output_loaded"]
            )
            if (
                st.session_state["segments"] is not None
                or len(st.session_state.segments) > 0
            ):
                if emd.n_variants > 2:
                    st.session_state.wp_segmentation_treatment = st.radio(
                        "Treatment to analyse",
                        np.arange(emd.n_variants - 1) + 1,
                    )
                else:
                    st.session_state.wp_segmentation_treatment = 1
                st.selectbox("Target", emd.targets, key="wp_target")

                st.dataframe(
                    segmentation_output.wise_pizza_output(
                        st.session_state.wp_target,
                        st.session_state.wp_segmentation_treatment,
                    )
                    .style.apply(highlight, axis=1)
                    .set_table_attributes("style='display:inline'")
                    .bar(subset=["Estimated_Effect_relative"], color="grey")
                    .format(precision=3)
                )
                wp_fig = segmentation_output.wise_pizza_figs[
                    st.session_state.wp_target
                ][st.session_state.wp_segmentation_treatment]
                st.plotly_chart(wp_fig, use_container_width=True)
                st.write("Expand figure below to fullscreen for best view.")

                st.markdown(
                    """
                    #### How can I interpret the results?

                    - We are trying to find unusual segments in terms of the averages
                    (***to highlight the segments contributing
                    the most to the difference between test and control***)
                    - Impact (Blue) is the model coefficient * size, if it is bigger
                    than zero, than segment average is bigger than the global one
                    (global difference in metric between test and control)
                    - Simple segment averages (Red) - average for specific segment
                    - Segment Sizes (Green) - number if observations in the segment
                    """
                )
            else:
                st.write("Select segments to analyse.")

            st.divider()
            st.subheader("Evaluation for segments")
            st.write(
                "In this section you can get p-value and statistical resuls for the"
                " selected segment"
            )

        if (
            st.session_state["segments"] is None
            or len(st.session_state["segments"]) == 0
        ):
            st.write("No segments available for frequentist segmentation analysis")
        else:
            st.markdown(
                """
                #### How can I interpret the results?

                - For each variant we compare it with the control group

                - Here we have mean values for every specified metric.
                - Then we have estimated effects and p-value for them
                - By default if p-value < 0.05 the result is significant
                - And then we have confidence intervals for metrics
                """
            )
            st.selectbox(
                "Segment to analyse",
                st.session_state["segments"],
                key="freq_eval_segments",
            )
            if len(st.session_state.freq_eval_segments) > 0:
                s = (
                    segmentation_output.segment_output(
                        st.session_state.freq_eval_segments, alpha=alpha / 100
                    )
                    .reset_index(level=[st.session_state.freq_eval_segments])
                    .style
                    # .apply(highlight, axis=1)
                    .set_table_attributes("style='display:inline'")
                    # .bar(subset=["Estimated_Effect_relative"], color="grey")
                    .format(precision=3)
                )
                st.dataframe(s)
    elif not st.session_state["is_experiment"]:
        st.write(
            """You have only provided pre-experiment data.
            Please define the experiment in "Experiment Design"
            first to use Experiment Evaluation.
            """
        )

    else:
        st.write("You first need to Choose or Configure the experiment..")

    st.session_state["last_page"] = CURRENT_PAGE



import plotly.express as px
from enum import Enum

USERNAME = "snowflake_username"  # your username
ACCOUNT = "account"
REGION = "region"
AUTHENTICATOR = "EXTERNALBROWSER"
DATABASE = "db"
WAREHOUSE = "warehouse"
# path to table with final results
RESULT_DATABASE = "SANDBOX_DB"
# path to table with final results
RESULT_SCHEMA = "sandbox_firstname_lastname"
# path to table with final results
RESULT_TABLE = "result_table"
# path to table with user_id, timestamp
SOURCE_DATABASE = "source_db"
# path to table with user_id, timestamp
SOURCE_SCHEMA = "source_schema"
# path to table with user_id, timestamp
SOURCE_TABLE = "source_table"
# user_id or profile_id column name (if profile_id, then set is_profile_id=True)
ID_COLUMN = "id_column"
# timestamp column name
TIMESTAMP_COLUMN = "timestamp_column"

WISE_COLORS = dict(
    dark_green="#163300",
    bright_orange="#ffc091",
    bright_blue="#a0e1e1",
    bright_yellow="#ffeb69",
    dark_purple="#ca7ce3",  # '#260a2f',
    dark_charcoal="#21231d",
    dark_gold="#ae9d58",  # '#3a341c',
    bright_pink=" #ffd7ef",
    dark_maroon="#320707",
    bright_light_green="#9fe870",
)

PLOTLY_COLOR_PALETTE = list(WISE_COLORS.values()) + px.colors.qualitative.G10
COLORSCALES = ["Blues", "Greens", "Purples", "Reds", "Oranges", "Greys", "BuGn"]


class MetricType(Enum):
    BINARY = "binary"
    CONTINUOUS = "continuous"
    DISCRETE = "discrete"



from econml.dr import LinearDRLearner
from tw_experimentation.variance_reduction.variance_reduction_method import (
    VarianceReductionMethod,
)
import pandas as pd
from typing import List
from sklearn.ensemble import HistGradientBoostingRegressor

# import warnings filter
from warnings import simplefilter

# ignore all future warnings
simplefilter(
    action="ignore", category=FutureWarning
)  # NOTE: should actually be investigated


class DoublyRobustEstimator(VarianceReductionMethod):
    """Implements a Doubly Robust Estimator for Causal Inference."""

    def fit(
        self,
        data: pd.DataFrame,
        treatment_column: str,
        target_column: str,
        covariate_columns: List[str],
        model_propensity="auto",
        model_propensity_init_config={},
        model_regression=HistGradientBoostingRegressor,
        model_regression_init_config={},
        **kwargs,
    ):
        """Fit the Doubly Robust Estimator to the data.s.

        Args:
            data (pd.DataFrame): experiment data
            treatment_column (str): name of the column containing the treatment flags
            target_column (str): name of the column containing the target metric
            covariate_columns (List[str]): list of names of covariate columns

        Returns:
            DoublyRobustEstimator: self
        """

        target = data[target_column]
        treatment = data[treatment_column]
        covariates = data[covariate_columns]

        # initialize models
        if model_propensity != "auto":
            initialized_model_propensity = model_propensity(
                **model_propensity_init_config
            )
        else:
            initialized_model_propensity = model_propensity
        initialized_model_regression = model_regression(**model_regression_init_config)

        # fit Doubly Robust Learner
        self.estimator = LinearDRLearner(
            model_propensity=initialized_model_propensity,
            model_regression=initialized_model_regression,
        )
        self.estimator.fit(Y=target, T=treatment, X=covariates, W=None)

        # fit a difference-in-means estimator
        self.fit_baseline(
            data=data, treatment_column=treatment_column, target_column=target_column
        )

        # generate statistics
        self.estimate = self.estimator.ate(
            X=covariates
        )  # TODO: should I use .effect(X=covariates) instead?
        self.conf_int_95 = self.conf_int(
            data=data, covariate_columns=covariate_columns, alpha=0.05
        )

        return self

    def calculate_variance_reduction():
        raise NotImplementedError("No explicit form, estimate through bootstrapping")

    def conf_int(self, data, covariate_columns, alpha: float = 0.05):
        """Calculate the 1-`alpha` * 100% confidence interval for the ATE estimate."""

        return self.estimator.ate_interval(
            X=data[covariate_columns].to_numpy(), alpha=alpha
        )

    def ci_width(self, data, covariate_columns, alpha: float = 0.05):
        """Calculate the width of the 1-`alpha` * 100% confidence interval."""

        ci = self.conf_int(data, covariate_columns, alpha=alpha)
        return ci[1] - ci[0]

    def calculate_ci_width_reduction(
        self, data, covariate_columns, alpha: float = 0.05
    ):
        """Calculate the width reduction rate in 1-`alpha` * 100% confidence
        interval."""

        baseline_conf_int = self.baseline_results.conf_int(alpha=alpha, cols=None)
        baseline_ci_width = baseline_conf_int[1][1] - baseline_conf_int[0][1]
        dr_ci_width = self.ci_width(data, covariate_columns, alpha=alpha)

        return 1 - (dr_ci_width / baseline_ci_width)



import streamlit as st
from streamlit_utils import initalise_session_states

st.session_state.update(st.session_state)
st.set_page_config(layout="wide")


CURRENT_PAGE = "Main"


initalise_session_states()

st.session_state["last_page"] = CURRENT_PAGE


st.title("TW Experimentation: AB Testing")

st.markdown(
    """
    ### How to use it
    - All images that appear dynamic have a download option. If you hover over the top right corner of the image, click on the camera symbol ("Download plot as a png")
    - If you would like to copy a displayed table, you can simple select all the relevant entries and copy-paste them to outside of the application.
    Note however that it is currently not possible to also copy the column names :(
    - Some tables do not display nicely immediately. You can expand them to full-screen on the top right corner of the table.
    ### What can this do for you?

    The experimentation library can help you with:
    - Sample Size Calculator
    - Integrity Checks
    - Evaluation (Frequentist A/B testing and Wise-pizza)
    - Evaluation (Bayesian A/B testing)

    #### 1. Designing experiments
    By using **TW Experimentation** you can design your experiments, choose sample size, evaluate the experiment and calculate features and metrics.


    #### 2. Evaluating results
    You can use different statistical tests and causal inference techniques.
    """
)



import os
from dataclasses import dataclass
from typing import List, Union, Optional
import pandas as pd
import numpy as np
from numpy.distutils.misc_util import is_sequence
from itertools import repeat

from tw_experimentation.constants import PLOTLY_COLOR_PALETTE, MetricType


def highlight(df):
    """Highlight significant results in green, non-significant in red for frequentist
    stat table."""
    if df["is_significant"]:
        if df["Estimated_Effect_relative"] > 0:
            return ["background-color: lightgreen"] * len(df)
        else:
            return ["background-color: lightcoral"] * len(df)
    else:
        return [""] * len(df)


def hex_to_rgb(hex_color: str) -> tuple:
    """Convert hex color format to rgb
    Taken from
    https://community.plotly.com/t/scatter-plot-fill-with-color-how-to-set-opacity-of-fill/29591/2
    Args:
        hex_color (str): hex color in format '#rrggbb'

    Returns:
        tuple: rgb color as tuple
    """
    hex_color = hex_color.lstrip("#")
    if len(hex_color) == 3:
        hex_color = hex_color * 2
    return int(hex_color[0:2], 16), int(hex_color[2:4], 16), int(hex_color[4:6], 16)


def variant_name_map(n_variants: int):
    """Name variants to be used in plots.

    Args:
        n_variants (int): Number of variants, includes control

    Returns:
        dict: map from variant number to variant name
    """
    assert n_variants > 0
    names = [
        f"Variant {i}" + (" (treatment)" if i > 0 else " (control)")
        for i in range(n_variants)
    ]

    name_map = dict(zip(range(n_variants), names))
    return name_map


def variant_color_map(n_variants: int):
    assert n_variants > 0
    color_map = dict(zip(range(n_variants), PLOTLY_COLOR_PALETTE))
    return color_map


def variantname_color_map(variant_names: List[str]):
    # """Map variant names to colors for plotting"""
    assert len(variant_names) > 0
    color_map = dict(zip(variant_names, PLOTLY_COLOR_PALETTE))
    return color_map


def create_dir(loc: str):
    parent = os.path.realpath(os.path.join(loc, ".."))
    if os.path.isdir(parent):
        if not os.path.isdir(loc):
            print("creating", loc)
            os.mkdir(loc)
    else:
        create_dir(parent)
        create_dir(loc)


@dataclass
class ExperimentMetaData:
    variant: str
    targets: List[str]
    n_variants: int
    date: Optional[str] = None
    pre_experiment_cols: Optional[List[str]] = None
    segments: Optional[List[str]] = None
    metric_types: Optional[dict[str, str]] = None
    is_dynamic_observation: Optional[bool] = None
    is_only_pre_experiment: Optional[bool] = False
    variant_labels: Optional[List[str]] = None
    variant_names: Optional[List[str]] = None
    sample_sizes: Optional[dict[str, int]] = None
    total_sample_size: Optional[int] = None
    target_standard_deviations: Optional[dict[str, float]] = None


@dataclass
class ExperimentDataset:
    data: pd.DataFrame
    variant: str
    targets: List[str]
    date: str
    ratio_targets: List[tuple]  # tuples column names for ratio metric calculation
    n_variants: Optional[int]

    def __init__(
        self,
        data: pd.DataFrame,
        variant: str,
        targets: Union[str, List[str]],
        pre_experiment_cols: Optional[List[str]] = None,
        segments: Optional[List[str]] = None,
        metric_types: Optional[dict[str, str]] = None,
        date: Optional[str] = None,
        ratio_targets: Optional[dict[str, tuple]] = None,
        n_variants: Optional[int] = 2,
        control_label: Optional[str] = 0,
        is_dynamic_observation: Union[Optional[bool], None] = None,
        is_only_pre_experiment: Optional[bool] = False,
    ) -> None:
        """Implements data logic for A/B testing. Assumes that observations are already
        on experiment analysis level.

        Args:
            data (pd.DataFrame): data with columns variant, target, date
            variant (str): variant column name
            targets (Union[str, List[str]]): target column name(s)
            pre_experiment_cols (Optional[List[str]], optional):
                pre-experimental data columns. Defaults to None.
            metric_types (Optional[dict[str, str]], optional):
                metric types ('binary', 'discrete', or 'continuous'). Defaults to None.
            date (Optional[str], optional): timestamp column. Defaults to None.
            ratio_targets (Optional[dict[str, tuple]], optional):
                ratio targets with numerator and denominator in tuple.
                Not implemented yet. Defaults to None.
            n_variants (Optional[int], optional): Number of variants
                TODO: autodetect this. Defaults to 2.
            control_label (Optional[str], optional): Label of the control group variant.
                Defaults to 0.
            is_dynamic_observation (Optional[bool], optional):
                Whether the assignment are dynamic
                (for monitoring and sequential analysis). Defaults to True.
            is_only_pre_experiment (Optional[bool], optional):
                Whether it only uses pre-experimental data.
        """
        # e.g. ratio_targets {'volume_per_transaction':
        # (total_volumes_per_customer, n_transactions_per_customer)}
        # will calculated mean and variance, covariance based on these stats

        assert isinstance(data, pd.DataFrame)
        self.data = data.copy()
        self.metric_types = metric_types
        self.date = date
        self.ratio_targets = ratio_targets
        self.n_variants = n_variants
        self.segments = segments or []
        self.control_label = control_label
        self.pre_experiment_cols = pre_experiment_cols or []
        self.is_dynamic_observation = is_dynamic_observation
        self.is_only_pre_experiment = is_only_pre_experiment
        self.variant = variant

        if not self.is_only_pre_experiment:
            assert isinstance(
                variant, str
            ), "Only a single treatment column supported at the moment"
            assert is_sequence(targets)
            self.n_variants = self.data[self.variant].nunique()

        self.targets = [targets] if isinstance(targets, str) else targets

        self.is_preprocessed = False

    def preprocess_pre_experiment_dataset(self):
        """
        - detec metric types
        """
        assert (
            self.is_only_pre_experiment is True
        ), "Can only preprocess data from pre-experiment"

        self._detect_metric_types(self.pre_experiment_cols)
        self._nan_data_cleaning()
        self._binary_metric_value_check()

    def preprocess_dataset(self, remove_outliers=True):
        """
        - calculate ratio metrics
        - remove outliers (optionally)
        """
        assert (
            self.is_only_pre_experiment is False
        ), "Can only preprocess data from experiment"
        # TODO: check if variant = 0,1[,2,...] or string.
        # If the latter, ask for name of control group
        # For now, assume that control = 0, treatment = 1,2,3,...

        # TODO: Convert date column to datetime
        self.data[self.variant] = self.data[self.variant].astype(int)
        try:
            assert 0 in list(self.data[self.variant].unique()) and 1 in list(
                self.data[self.variant].unique()
            )
        except Exception as e:  # noqaF841
            raise Exception("Variants must be 0 (=control), 1,[2,...] (=treatment)")

        experiment_cols = [
            self.targets,
            [self.variant],
            self.segments,
            self.pre_experiment_cols,
        ]
        if self.date is not None:
            experiment_cols.append([self.date])

        assert all(
            (
                set(col_type).issubset(set(self.data.columns)),
                "Columns must be in dataframe",
            )
            for col_type in experiment_cols
        )

        experiment_cols_list = [col for cols in experiment_cols for col in cols]
        assert len(set(experiment_cols_list)) == len(
            experiment_cols_list
        ), "Columns must be unique"

        if self.metric_types is None:
            self.metric_types = dict(zip(self.targets, repeat(None)))
        missing_metric_types = [
            t for t, v in self.metric_types.items() if v is None and t in self.targets
        ]
        if self.pre_experiment_cols is not None and len(self.pre_experiment_cols) > 0:
            missing_metric_types += self.pre_experiment_cols

        self._nan_data_cleaning()
        self._detect_metric_types(missing_metric_types)
        available_types = [metric_type.value for metric_type in MetricType]
        assert all(
            metric_type in available_types for metric_type in self.metric_types.values()
        ), "Metric types must be one of 'binary', 'discrete', or 'continuous'"

        self._binary_metric_value_check()

        if remove_outliers:
            self._remove_outliers()

        if self.is_dynamic_observation is None:
            self._detect_is_dynamic_observation()

        self.is_preprocessed = True

    def _binary_metric_value_check(self):
        assert all(
            self.data[metric].isin([0, 1]).all()
            for metric, metric_type in self.metric_types.items()
            if metric_type == MetricType.BINARY.value
        ), "Binary metric types must consist of values 0 or 1"

    def _detect_is_dynamic_observation(self):
        """Detect whether the assignment of variants is dynamic (for monitoring and
        sequential analysis)"""
        if self.date is not None and self.data[self.date].nunique() > 1:
            self.is_dynamic_observation = True
        else:
            self.is_dynamic_observation = False

    def _nan_data_cleaning(self):
        self.data = self.data.fillna(0)

    def _detect_metric_types(self, metrics):
        """Detect metric types with the following logic:
            if integer:
                check whether binary 0,1
                    yes -> 'binary'
                    no -> 'discrete
            elif other type:
                try converting to float
                if binary -> 'binary'
                elif float is integer -> 'discrete'
                else -> 'continuous'
            else raise error

        Args:
            metrics (List[str]): List of metrics for which to detect metric type
                Must be columns in self.data

        Raises:
            TypeError: If metrics cannot be converted to number, raise error

        """
        assert set(metrics).issubset(self.data.columns)
        if self.metric_types is None:
            self.metric_types = {}
        for metric in metrics:
            error_msg = f"Metrics must be numbers, but column {metric} is not."
            auto_dtype = self.data[metric].dtype
            if isinstance(self.data[metric].dtype, int):
                if set(self.data[metric].values) == set([0, 1]):
                    self.metric_types[metric] = MetricType.BINARY.value
                else:
                    self.metric_types[metric] = MetricType.DISCRETE.value
            elif isinstance(auto_dtype, (float, str, object)):
                try:
                    self.data[metric] = self.data[metric].fillna(np.nan).apply(float)
                    if self.data[metric].apply(float.is_integer).all():
                        self.data[metric].fillna(np.nan).apply(int)
                        if set(self.data[metric].values) == set((0, 1)):
                            self.metric_types[metric] = MetricType.BINARY.value
                        else:
                            self.metric_types[metric] = MetricType.DISCRETE.value
                    else:
                        self.metric_types[metric] = MetricType.CONTINUOUS.value
                except Exception as TypeError:  # noqaF841
                    print(error_msg)
            else:
                raise TypeError(error_msg)

    def _aggregator(self):
        # Similar as aggregate preprocessor in
        # https://github.com/MobileTeleSystems/Ambrosia/blob/main/examples/1_usage_example_for_core_classes.ipynb
        pass

    def _remove_outliers(self):
        outliers_dict = {}
        for target in self.targets:
            if self.metric_types[target] == MetricType.CONTINUOUS.value:
                outliers_dict[target] = np.percentile(self.data[target], 99)
        if len(outliers_dict) > 0:
            query_string = " and ".join(
                [
                    f"{target}<={perc}"
                    for (target, perc) in zip(
                        outliers_dict.keys(), outliers_dict.values()
                    )
                ]
            )
            self.data = self.data.query(query_string)

    def _diff_in_diff(self):
        # to remove pre-experimental bias
        pass

    def experiment_meta_data(self):
        emd = ExperimentMetaData(
            variant=self.variant,
            targets=self.targets,
            n_variants=self.n_variants,
            date=self.date,
            pre_experiment_cols=self.pre_experiment_cols,
            segments=self.segments,
            metric_types=self.metric_types,
            is_dynamic_observation=self.is_dynamic_observation,
            is_only_pre_experiment=self.is_only_pre_experiment,
            variant_labels=self.variant_labels,
            variant_names=self.variant_names,
            sample_sizes=self.sample_sizes,
            total_sample_size=self.total_sample_size,
            target_standard_deviations=self.target_standard_deviations,
        )
        return emd

    @property
    def target_standard_deviations(self):
        """Compute standard deviation of each target for control group."""
        return {
            target: self.data.loc[self.data[self.variant] == 0, target].std()
            for target in self.targets
        }

    @property
    def variant_labels(self):
        return np.sort(self.data[self.variant].unique())

    @property
    def get_metric_types(self):
        return self.metric_types

    @property
    def variant_names(self):
        # TODO: Check whether this is duplicated from variant_labels
        return variant_name_map(self.n_variants)

    @property
    def sample_sizes(self):
        return {
            label: len(self.data[self.data[self.variant] == label].index)
            for label in self.variant_labels
        }

    @property
    def total_sample_size(self):
        return len(self.data)



from tw_experimentation.streamlit.streamlit_utils import (
    bayes_cache_wrapper,
    initalise_session_states,
)
import streamlit as st
import plotly.graph_objects as go


def page_5_experiment_evaluation_bayesian():
    st.session_state.update(st.session_state)

    CURRENT_PAGE = "Bayesian Evaluation"
    initalise_session_states()

    st.header("Bayesian AB test evaluation")

    if st.session_state["is_defined_data_model"] and st.session_state["is_experiment"]:
        if st.session_state["output_loaded"] is not None:
            emd = st.session_state["output_loaded"]["emd"]
        else:
            emd = st.session_state.ed.experiment_meta_data()
        st.write(
            "In this section, we fit a Bayesian model on the data."
            "We then use the posterior"
            " distribution to compute the probability that the variant "
            "is better than the control."
        )

        br = bayes_cache_wrapper(
            st.session_state.ed, _output_loaded=st.session_state["output_loaded"]
        )

        st.selectbox("Target", emd.targets, key="bayes_target_plot")

        fig = br.fig_posterior_difference_by_target(
            st.session_state["bayes_target_plot"]
        )
        st.plotly_chart(fig)

        fig = br.fig_posterior_cdf_by_target(st.session_state["bayes_target_plot"])

        st.plotly_chart(fig)

        # TODO: Enable bayes factor when tested properly
        # st.subheader("Bayes Factor")

        description = """
            The Bayes Factor is the p-value analogue in Bayesian hypothesis testing.
            It allows us to compare the hypothesis of no effect (null hypothesis)
            to the hypothesis of that there is an effect (alternative hypothesis).
            The decision is made based on the Risk, which is the probability that
            the null hypothesis is true given the data
            (probability of a false discovery)."""
        # TODO: Enable bayes factor when tested properly
        # st.write(description)
        fdrs = [
            br.false_discovery_rate(st.session_state["bayes_target_plot"], var)
            for var in range(1, st.session_state.ed.n_variants)
        ]
        fdrs = [f"{fdr:.2f}" for fdr in fdrs]
        bayes_factors = [
            f'{br.bayes_factor(st.session_state["bayes_target_plot"], var):.2f}'
            for var in range(1, st.session_state.ed.n_variants)
        ]
        fig = go.Figure(
            data=[
                go.Table(
                    header=dict(
                        values=[
                            "<b>Bayes Factor</b>",
                            "<b>Risk (%)</b>",
                            "<b>Decision</b>",
                        ],
                        align="left",
                    ),
                    cells=dict(
                        values=[
                            bayes_factors,
                            fdrs,
                            [
                                br.bayes_factor_decision(
                                    st.session_state["bayes_target_plot"], var
                                )
                                for var in range(1, st.session_state.ed.n_variants)
                            ],
                        ],
                        align="left",
                    ),
                )
            ]
        )
        # TODO: Enable bayes factor when tested properly
        # st.plotly_chart(fig)
        st.subheader("Probability that treatment effect greater or smaller than 0")

        n_variants = emd.n_variants
        n_cols_max = 3
        row_order = n_variants <= n_cols_max + 1

        description = (
            "The probability that the average treatment effect is"
            "**greater than 0** for"
            f" outcome metric **{st.session_state['bayes_target_plot']}** is\n"
        )
        st.write(description)

        if row_order:
            cols = st.columns(n_variants - 1)

        for variant in range(1, n_variants):
            if row_order:
                with cols[variant - 1]:
                    prob_greater_zero = (
                        br.prob_greater_than_zero(
                            st.session_state["bayes_target_plot"]
                        )[variant]
                        * 100
                    )
                    st.metric(
                        f"{emd.variant_names[variant]}",
                        value=(f"{prob_greater_zero:.2f} %"),
                    )
            else:
                prob_greater_zero = (
                    br.prob_greater_than_zero(st.session_state["bayes_target_plot"])[
                        variant
                    ]
                    * 100
                )
                st.metric(
                    f"{emd.variant_names[variant]}",
                    value=(f"{prob_greater_zero:.2f} %"),
                )

        description = (
            "The probability that the average treatment effect"
            "is **smaller than 0** for"
            f" outcome metric **{st.session_state['bayes_target_plot']}** is\n"
        )
        st.write(description)

        if row_order:
            cols = st.columns(n_variants - 1)

        for variant in range(1, n_variants):
            if row_order:
                with cols[variant - 1]:
                    prob_smaller_zero = (
                        1
                        - br.prob_greater_than_zero(
                            st.session_state["bayes_target_plot"]
                        )[variant]
                    ) * 100
                    st.metric(
                        f"{emd.variant_names[variant]}",
                        value=(f"{prob_smaller_zero:.2f} %"),
                    )
            else:
                prob_smaller_zero = (
                    1
                    - br.prob_greater_than_zero(st.session_state["bayes_target_plot"])[
                        variant
                    ]
                ) * 100
                st.metric(
                    f"{emd.variant_names[variant]}",
                    value=(f"{prob_smaller_zero:.2f} %"),
                )

        st.subheader(
            "Probability that treatment effect is above or below a custom threshold"
        )
        st.write(
            """

            The probability that the average treatment effect is greater
            or smaller than a custom `threshold`:
            """
        )

        st.number_input(
            "threshold",
            key="bayes_threshold",
        )

        description = (
            "The probability that the average treatment effect is greater than"
            f" **{st.session_state['bayes_threshold']}** for outcome metric"
            f" **{st.session_state['bayes_target_plot']}** is\n"
        )
        st.write(description)

        if row_order:
            cols2 = st.columns(n_variants - 1)

        for variant in range(1, n_variants):
            if row_order:
                with cols2[variant - 1]:
                    prob_greater_z = (
                        br.prob_greater_than_z(
                            st.session_state["bayes_threshold"],
                            st.session_state["bayes_target_plot"],
                        )[variant]
                        * 100
                    )
                    st.metric(
                        f"{emd.variant_names[variant]}",
                        value=(f"{prob_greater_z:.2f} %"),
                    )
            else:
                prob_greater_z = (
                    br.prob_greater_than_z(
                        st.session_state["bayes_threshold"],
                        st.session_state["bayes_target_plot"],
                    )[variant]
                    * 100
                )
                st.metric(
                    f"{emd.variant_names[variant]}",
                    value=(f"{prob_greater_z:.2f} %"),
                )

            description = (
                "The probability that the average treatment effect is smaller than"
                f" **{st.session_state['bayes_threshold']}** for outcome metric"
                f" **{st.session_state['bayes_target_plot']}** is\n"
            )
        st.write(description)

        if row_order:
            cols2 = st.columns(n_variants - 1)

        for variant in range(1, n_variants):
            if row_order:
                with cols2[variant - 1]:
                    prob_smaller_z = (
                        br.prob_smaller_than_z(
                            st.session_state["bayes_threshold"],
                            st.session_state["bayes_target_plot"],
                        )[variant]
                        * 100
                    )
                    st.metric(
                        f"{emd.variant_names[variant]}",
                        value=(f"{prob_smaller_z:.2f} %"),
                    )
            else:
                prob_smaller_z = (
                    br.prob_smaller_than_z(
                        st.session_state["bayes_threshold"],
                        st.session_state["bayes_target_plot"],
                    )[variant]
                    * 100
                )
                st.metric(
                    f"{emd.variant_names[variant]}",
                    value=(f"{prob_smaller_z:.2f} %"),
                )

        st.subheader(
            "Probability that treatment effect is in ROPE"
            "(Region of Practical Equivalence)"
        )

        st.markdown(
            """
            As an analogue to frequentist testing,
            you can define a region of practical equivalence (ROPE).
            This is an interval where we assume that when the effect is
            within the interval,
            the effect is negligible
            (e.g. not big enough to make an impact,
            cover cost of rolling out the change).

            Another reason for defining a ROPE interval is that when we simply
            look at the
            probability that the effect is greater than 0, that probability
            will be 50% even without using any data if we assume that the effect
            is centered around 0.
            The ROPE interval sizes are autodetected based on the variance
            of the outcome metric.
            """
        )
        probs, rope_lower, rope_upper = br.rope(st.session_state["bayes_target_plot"])
        description = (
            "The probability that the average treatment effect is"
            "outside the **region of"
            f" practical equivalence ({rope_lower:.2f}, {rope_upper:.2f})** for outcome"
            f" metric **{st.session_state['bayes_target_plot']}** is\n"
        )
        st.write(description)

        if row_order:
            cols3 = st.columns(n_variants - 1)

        for variant in range(1, emd.n_variants):
            if row_order:
                with cols3[variant - 1]:
                    st.metric(
                        f"{emd.variant_names[variant]}",
                        value=f"{probs[variant]*100:.2f} %",
                    )
            else:
                st.metric(
                    f"{emd.variant_names[variant]}",
                    value=f"{probs[variant]*100:.2f} %",
                )

        st.subheader("Probability that treatment effect is outside a custom interval")

        col_input_lower, col_input_upper = st.columns(2)
        with col_input_lower:
            st.number_input(
                "Lower threshold",
                key="bayes_threshold_lower",
            )
        with col_input_upper:
            st.number_input(
                "Upper threshold",
                key="bayes_threshold_upper",
            )

        description = (
            "The probability that the average treatment effect is"
            "**outside the interval"
            f" ({st.session_state['bayes_threshold_lower']:.2f},"
            f" {st.session_state['bayes_threshold_upper']:.2f})** for outcome metric"
            f" **{st.session_state['bayes_target_plot']}** is\n"
        )
        st.write(description)

        if row_order:
            cols4 = st.columns(n_variants - 1)

        probs = br.prob_outside_interval(
            st.session_state["bayes_threshold_lower"],
            st.session_state["bayes_threshold_upper"],
            st.session_state["bayes_target_plot"],
        )

        for variant in range(1, emd.n_variants):
            if row_order:
                with cols4[variant - 1]:
                    st.metric(
                        f"{emd.variant_names[variant]}",
                        value=f"{probs[variant]*100:.2f} %",
                    )
            else:
                st.metric(
                    f"{emd.variant_names[variant]}",
                    value=f"{probs[variant]*100:.2f} %",
                )
    elif not st.session_state["is_experiment"]:
        st.write(
            "You have only provided pre-experiment data. "
            'Please define the experiment in "Experiment Design" first to '
            "use Experiment Evaluation Bayesian."
        )

    else:
        st.write("You first need to Choose or Configure the experiment.")

    st.session_state["last_page"] = CURRENT_PAGE



"""
Tool to set up experiments: alpha, beta, MDE, power
Code contains Setuper class with methods.
"""

import numpy as np

from dataclasses import dataclass

import statsmodels.stats as stats
from statsmodels.stats import power as pwr

import plotly.express as px

from tw_experimentation.utils import ExperimentDataset


@dataclass
class ExpDesignAutoCalculate:
    """A class for automatically calculating standard deviation and mean values for pre-
    experiment columns in an ExperimentDataset.

    Attributes:
        ed (ExperimentDataset): The ExperimentDataset object containing
            the pre-experiment data.
    """

    ed: ExperimentDataset

    def __post_init__(self):
        self.sds = {
            target_pre_exp: self.ed.data[target_pre_exp].std()
            for target_pre_exp in self.pre_experiment_cols
        }

        self.means = {
            target_pre_exp: self.ed.data[target_pre_exp].mean()
            for target_pre_exp in self.pre_experiment_cols
        }

    @property
    def metric_types(self):
        return self.ed.metric_types

    @property
    def pre_experiment_cols(self):
        """
        Returns:
            List[str]: A list of pre-experiment column names.
        """
        return self.ed.pre_experiment_cols

    def sd(self, target_pre_exp):
        """
        Args:
            target_pre_exp (str): The name of the pre-experiment column.

        Returns:
            float: The standard deviation value for the given pre-experiment column.
        """
        return self.sds[target_pre_exp]

    def mean(self, target_pre_exp):
        """
        Args:
            target_pre_exp (str): The name of the pre-experiment column.

        Returns:
            float: The mean value for the given pre-experiment column.
        """
        return self.means[target_pre_exp]


class Setuper:
    """Tool for designing AB tests.

    Result includes:
        - Minimal Detectable Effect Size
        - Sample Size calculation
        - Sample Size vs MDE plot
    """

    def __init__(
        self,
        alpha=0.05,
        beta=0.2,
        treatment_share=0.5,
        effect_size=0.02,
        alternative="two-sided",
    ) -> None:
        """

        Args:
            alpha (float, optional): Type-I error. Defaults to .05.
            beta (float, optional): Type-II error. Defaults to .2.
            treatment_share (float, optional): Share of treatment group in the sample.
                Defaults to .5.
            effect_size (float, optional): Standardized effect size: Cohen d.
                Defaults to .02.
            alternative (str, optional): 'two_sided', 'larger', 'smaller'
            for alternative test designs. Defaults to 'two-sided'.
        """

        self.alpha = alpha
        self.beta = beta
        self.treatment_share = treatment_share
        self.effect_size = effect_size
        self.ratio = treatment_share / (1 - treatment_share)
        self.alternative = alternative

    @classmethod
    def from_uplift(
        cls,
        uplift,
        sd,
        mean,
        alpha=0.05,
        beta=0.2,
        treatment_share=0.5,
        relation="absolute",
    ):
        """Design test size from uplift instead of standardized effect size.

        Args:
            uplift (float): uplift to detect
            sd (float): standard deviation
            mean (float): expected baseline mean
            alpha (float, optional): type-I error. Defaults to .05.
            beta (float, optional): type-II error. Defaults to .2.
            treatment_share (float, optional): treatment share in population.
                Defaults to .5.
            relation (str, optional): 'two-sided', 'larger' or 'smaller.
                Defaults to 'absolute'.

        Returns:
            Setuper class instance: Setuper class with uplift
                converted to effect size
        """
        return cls(
            alpha,
            beta,
            treatment_share,
            uplift_to_effect_size(uplift, mean, sd, relation),
        )

    def sample_size_t_test(self):
        """T test sample size calculation for continuous outcomes.

        Returns:
            dict: sample sizes per group
        """
        hyp_test = getattr(stats.power, "tt_ind_solve_power")
        treat_control_ratio = (1 - self.treatment_share) / self.treatment_share
        treatment_ss = hyp_test(
            effect_size=self.effect_size,
            alpha=self.alpha,
            power=1 - self.beta,
            ratio=treat_control_ratio,
        )
        total_ss = int(treatment_ss) + int(treatment_ss + treat_control_ratio)
        return {
            "Total Sample Size": int(total_ss),
            "Treatment Sample Size": int(treatment_ss),
            "Control Sample Size": int(treatment_ss * treat_control_ratio),
        }

    def sample_size_two_sample_proportion_z_test(self):
        """Two sample proportion z test sample size calculation.

        Returns:
            dict: sample sizes per group
        """
        hyp_test = getattr(stats.power, "zt_ind_solve_power")
        treat_control_ratio = (1 - self.treatment_share) / self.treatment_share
        treatment_ss = hyp_test(
            effect_size=self.effect_size,
            alpha=self.alpha,
            power=1 - self.beta,
            ratio=treat_control_ratio,
        )
        total_ss = treatment_ss * (1 + treat_control_ratio)
        return {
            "Total Sample Size": int(total_ss),
            "Treatment Sample Size": int(treatment_ss),
            "Control Sample Size": int(treatment_ss * treat_control_ratio),
        }

    def sample_size_chi_squared_test(self):
        """Sample Size calculation for chi squared test
        statsmodels.stats.proportion.proportions_chisquare."""
        pass

    def sample_size_proportion_test(
        self, uplift, baseline_conversion=0.5, relation="absolute"
    ):
        """Sample sie for proportion test, requires uplift and
            automatically calculate standard deviation.

        Args:
            uplift (float): uplift to detect
            baseline_conversion (float, optional): baseline conversion.
                Defaults to .5.
            relation (str, optional): 'two-sided', 'larger' or 'smaller'.
                Defaults to 'absolute'.

        Returns:
            float: total sample size
        """

        if relation == "absolute":
            pass
        elif relation == "relative":
            uplift = uplift * baseline_conversion

        return (
            stats.proportion.samplesize_proportions_2indep_onetail(
                diff=uplift,
                prop2=baseline_conversion,
                power=1 - self.beta,
                alpha=self.alpha,
                alternative=self.alternative,
            )
            / self.treatment_share
        )

    def effect_size_proportion_test(self):
        """Effect size for statsmodels.stats.proportion.test_proportions_2indep."""
        pass

    def effect_size_t_test(self, nobs):
        """Effect size for t test given power and number of observations.

        Args:
            nobs (int): number of observations

        Returns:
            float: minimal detectable effect size
        """
        nobs1 = nobs * self.treatment_share
        ratio = self.ratio
        analysis = pwr.TTestIndPower()
        esresult = analysis.solve_power(
            power=1 - self.beta, nobs1=nobs1, ratio=ratio, alpha=self.alpha
        )
        return esresult

    def effect_size_two_sample_z_test(self, nobs):
        """Effect size of two sample independent proportion z test.

        Args:
            nobs (int): number of observations

        Returns:
            float: minimal detectable effect size
        """

        hyp_test = getattr(stats.power, "zt_ind_solve_power")
        nobs1 = nobs * self.treatment_share
        treat_control_ratio = (1 - self.treatment_share) / self.treatment_share
        mde = hyp_test(
            nobs1=nobs1,
            alpha=self.alpha,
            power=1 - self.beta,
            ratio=treat_control_ratio,
        )
        return mde

    def effect_size_chi_squared_test(self):
        """TODO: MDE for chi-squared test"""
        pass

    def power_t_test(self, nobs):
        """Power of t test for continuous outcomes.

        Args:
            nobs (int): number of observations

        Returns:
            float: power
        """
        ratio = self.ratio
        analysis = pwr.TTestIndPower()
        pwresult = analysis.solve_power(
            effect_size=self.effect_size,
            power=None,
            alpha=self.alpha,
            nobs1=nobs,
            ratio=ratio,
        )
        return pwresult

    def power_two_sample_z_test(self, nobs):
        """Power two sample independent proportion z test.

        Args:
            nobs (int): number of observations

        Returns:
            float: power
        """
        hyp_test = getattr(stats.power, "zt_ind_solve_power")
        nobs1 = nobs * self.treatment_share
        treat_control_ratio = (1 - self.treatment_share) / self.treatment_share
        power = hyp_test(
            nobs1=nobs1,
            alpha=self.alpha,
            effect_size=self.effect_size,
            ratio=treat_control_ratio,
        )
        return power

    def power_proportion_test(
        self, uplift, nobs, baseline_conversion=0.5, relation="absolute"
    ):
        """Power proportion test (as in evan miller)

        Args:
            uplift (float): uplift to detect
            nobs (int): number of observations
            baseline_conversion (float, optional): baseline conversion rate.
                Defaults to .5.
            relation (str, optional): 'two-sided', 'larger' or 'smaller'.
                Defaults to 'absolute'.

        Returns:
            float: power
        """

        if relation == "absolute":
            pass
        elif relation == "relative":
            uplift = uplift * baseline_conversion

        nobs1 = nobs * self.treatment_share

        return stats.proportion.power_proportions_2indep(
            diff=uplift,
            nobs1=nobs1,
            ratio=self.ratio,
            prop2=baseline_conversion,
            alpha=self.alpha,
            alternative=self.alternative,
        )

    def plot_sample_size_to_mde(
        self,
        effect_size_function,
        max_sample_size=1000,
    ):
        """Plot sample size to MDE.

        Args:
            effect_size_function (function): function that maps sample size to MDE
            max_sample_size (int, optional): x axis limit. Defaults to 1000.
        """
        # TODO: Also plot to sample size to uplift
        sample_size = np.arange(200, max_sample_size, 10)
        mde = np.array(list(map(effect_size_function, sample_size)))
        fig = px.line(
            x=sample_size,
            y=mde,
            width=600,
            height=600,
            title="MDE per Sample Size",
            labels={"x": "Sample Size", "y": "MDE"},
        )
        fig.show()


def uplift_to_effect_size(uplift, mean, sd, relation="absolute"):
    """Convert uplift to effect size.

    Args:
        uplift (float): expected uplift
        mean (float): estimated mean
        sd (float): estimated standard deviation

    Returns:
        float: effect size for power calculation
    """

    if relation == "absolute":
        es = uplift / sd
    elif relation == "relative":
        es = (uplift * mean) / sd
    else:
        raise Exception('Uplfit relation must be "relative" or "abosolut"')
    return es


def effect_size_to_uplift(es, mean, sd, relation="absolute"):
    """Convert effect size to uplift.

    Args:
        es (float): expected effect size
        mean (float): estimated mean
        sd (float): estimated standard deviation

    Returns:
        float: uplift for power calculation
    """
    if relation == "absolute":
        uplift = es * sd
    elif relation == "relative":
        uplift = es * (sd / mean)
    else:
        raise Exception('Uplfit relation must be "relative" or "abosolut"')
    return uplift



from tw_experimentation.variance_reduction.variance_reduction_method import (
    VarianceReductionMethod,
)
import pandas as pd


class DifferenceInMeans(VarianceReductionMethod):
    """Implement Difference-in-Means (DiM) estimator."""

    def fit(
        self, data: pd.DataFrame, treatment_column: str, target_column: str, **kwargs
    ):
        """Apply Difference In Means to data.

        Args:
            data (pd.DataFrame): experiment data containing pre-experiment data column
            treatment_column (str): name of column containing treatment flags
            target_column (str): name of column containing target metric

        Returns:
            DifferenceInMeans: self
        """

        # fit difference-in-means
        self.fit_baseline(
            data=data, treatment_column=treatment_column, target_column=target_column
        )

        # copy from baseline
        self.regression_results = self.baseline_results
        self.estimate = self.baseline_estimate
        self.p_value = self.baseline_p_value
        self.conf_int_95 = self.baseline_conf_int_95
        self.variance_reduction_rate = 0

        return self

    def calculate_variance_reduction(self):
        raise NotImplementedError

    def ci_width(self, alpha: float = 0.05):
        """Calculate the `1-alpha * 100%` confidence interval."""

        return (
            self.regression_results.conf_int(alpha=alpha, cols=None)[1][1]
            - self.regression_results.conf_int(alpha=alpha, cols=None)[0][1]
        )

    def calculate_ci_width_reduction(self, alpha=0.05):
        raise NotImplementedError
